#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
from datetime import date, datetime

from dateutil.tz import tzutc

from soql.nodes import ColumnPath, Expression, OPS, _stringify
from soql.nodes import and_, or_, not_
from tests.helpers import SoqlAssertions


class ExpressionTest(unittest.TestCase, SoqlAssertions):
    def test_operations(self):
        for operation in OPS.values():
            exp = Expression(lhs=123, op=operation, rhs=456)
            self.assertSoqlEqual(exp, '123 {} 456'.format(operation))

    def test_chained_expressions(self):
        exp1 = Expression(lhs=123, op=OPS.EQ, rhs=456)
        exp2 = Expression(lhs=789, op=OPS.AND, rhs=exp1)
        self.assertSoqlEqual(exp2, '789 AND 123 = 456')

    def test_type_rendering(self):
        exp1 = Expression(lhs='123', op=OPS.EQ, rhs=456)
        self.assertSoqlEqual(exp1, "'123' = 456")

        exp2 = Expression(lhs='789', op=OPS.AND, rhs=exp1)
        self.assertSoqlEqual(exp2, "'789' AND '123' = 456")


class StringifyTest(unittest.TestCase):
    def test_stringify(self):
        self.assertEqual(_stringify('Apples'), "'Apples'")

        self.assertEqual(_stringify(u'CATMONKÈ-123490'), u"'CATMONKÈ-123490'")

        self.assertEqual(_stringify(1), '1')

        self.assertEqual(_stringify(True), 'TRUE')

        self.assertEqual(_stringify(False), 'FALSE')

        self.assertEqual(_stringify(None), 'NULL')

        date_ = date(year=1990, month=10, day=19)
        self.assertEqual(_stringify(date_), '1990-10-19')

        datetime_ = datetime(
            year=1990,
            month=10,
            day=19,
            hour=1,
            minute=2,
            second=3,
            tzinfo=tzutc()
        )

        self.assertEqual(_stringify(datetime_), '1990-10-19T01:02:03+00:00')

        naive_datetime = datetime(
            year=1990,
            month=10,
            day=19,
            hour=1,
            minute=2,
            second=3
        )
        self.assertEqual(_stringify(naive_datetime), '1990-10-19T01:02:03+00:00')


class OperationsTest(unittest.TestCase):
    def test_operations(self):
        column_path = ColumnPath('Monkey.Tail')

        self.assertEqual(str(column_path == '123'), "Monkey.Tail = '123'")
        self.assertEqual(str(column_path == 56), "Monkey.Tail = 56")

        self.assertEqual(str(column_path < 5), "Monkey.Tail < 5")
        self.assertEqual(str(column_path > 5), "Monkey.Tail > 5")
        self.assertEqual(str(column_path <= 5), "Monkey.Tail <= 5")
        self.assertEqual(str(column_path >= 5), "Monkey.Tail >= 5")

        self.assertEqual(str(column_path != None), "Monkey.Tail != NULL")
        self.assertEqual(str(column_path == True), "Monkey.Tail = TRUE")
        self.assertEqual(str(column_path == False), "Monkey.Tail = FALSE")

        self.assertEqual(
            str(column_path.like_('Va%')),
            "Monkey.Tail LIKE 'Va%'"
        )
        self.assertEqual(
            str(column_path.in_(['Jin', 'Jan'])),
            "Monkey.Tail IN ('Jin', 'Jan')"
        )
        self.assertEqual(
            str(column_path.in_([u'Jin', u'Jan'])),
            "Monkey.Tail IN ('Jin', 'Jan')"
        )
        self.assertEqual(
            str(column_path.not_in_(['Jin', 'Jan'])),
            "Monkey.Tail NOT IN ('Jin', 'Jan')"
        )
        self.assertEqual(
            str(column_path.includes_(['Jin', 'Jan'])),
            "Monkey.Tail INCLUDES ('Jin', 'Jan')"
        )
        self.assertEqual(
            str(column_path.excludes_(['Jin', 'Jan'])),
            "Monkey.Tail EXCLUDES ('Jin', 'Jan')"
        )

        self.assertEqual(
            str(and_(column_path == 'R', column_path < 10)),
            "(Monkey.Tail = 'R' AND Monkey.Tail < 10)"
        )
        self.assertEqual(
            str(and_(column_path == 'R', column_path < 10, column_path != 'G')),
            "(Monkey.Tail = 'R' AND Monkey.Tail < 10 AND Monkey.Tail != 'G')"
        )

        self.assertEqual(
            str(or_(column_path == 'R', column_path < 10)),
            "(Monkey.Tail = 'R' OR Monkey.Tail < 10)"
        )
        self.assertEqual(
            str(or_(column_path == 'R', column_path < 10, column_path != 'G')),
            "(Monkey.Tail = 'R' OR Monkey.Tail < 10 OR Monkey.Tail != 'G')"
        )

        self.assertEqual(
            str(not_(column_path == 'R')),
            "NOT Monkey.Tail = 'R'"
        )

        self.assertEqual(
            str(not_(and_(column_path == 'R',
                          or_(column_path == 10, column_path == 'G')))),
            "NOT (Monkey.Tail = 'R' AND (Monkey.Tail = 10 OR Monkey.Tail = 'G'))"
        )



"""
    Model
    ~~~~~

    Base class for declarative models.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""

from __future__ import absolute_import
from collections import OrderedDict
from copy import copy

from six import with_metaclass

from soql.attributes import AttributeBase
from soql.attributes import Column
from soql.attributes import Relationship
from soql.model_registry import model_registry
from soql.path_builder import PathBuilder


class UNLOADED(object):
    """Placeholder for a model attribute value for data that was never loaded
    from the API."""
    pass


class UNSET(object):
    """Placeholder for a model attribute value for data that was never set."""
    pass


class AttributeNotLoaded(Exception):
    pass


class AttributeNotSet(Exception):
    pass


class ExpectedColumnMissing(Exception):
    pass


class _ModelMeta(type):
    """
    Turns boring classes into magical, declarative style models!
    """
    def __new__(mcs, name, bases, cls_attrs):
        """
        Metaclasses are classes for classes... which means this method is
        responsible for creating a new class. That means we can hijack the
        class declaration to do magical things...

        :param name: The name of the class being declared
        :param bases: The base classes the class is being declared with
        :param cls_attrs: A dictionary of attributes the class is being declared with
        :return: the new class
        """
        # This handles inheritance by crawling through all the base classes
        # and grabbing their declared attributes.
        inherited_attributes = OrderedDict()
        for base_ in bases:
            if hasattr(base_, 'declared_attrs'):
                inherited_attributes.update(
                    copy(getattr(base_, 'declared_attrs')))

        # Find any attributes that are declared in the model.
        # It helps testability tremendously if the attributes come in a reliable
        # order... hence the sorting and OrderedDict
        attributes = OrderedDict()
        for key in sorted(cls_attrs.keys()):
            if isinstance(cls_attrs[key], AttributeBase):
                attributes[key] = cls_attrs[key]

        # Pop them out of the attribute dictionary so they don't get declared
        # as normal attributes....
        for key, attr in attributes.items():
            cls_attrs.pop(key)

        # Now that the attributes have been pruned, we can create the class
        klass = super(_ModelMeta, mcs).__new__(mcs, name, bases, cls_attrs)

        # We'll attach those attributes, along with the attributes from the
        # inherited classes, to a single attribute.
        setattr(klass, 'declared_attrs', OrderedDict())
        klass.declared_attrs.update(inherited_attributes)
        klass.declared_attrs.update(attributes)

        # Finally, register the model in the model registry.
        if not cls_attrs.get('__abstract__'):
            model_registry.add(klass)

        return klass

    def __getattr__(self, item):
        """
        This little bit of magic helps us do things like:

            select(SomeModel).where(SomeModel.name == 'BLAM!')

        While still having instances of the model return the attribute's value:

            assert SomeModel().name == 'BLAM!'

        (Note - because this is a meta class, "self" is the model)
        """
        if item not in self.declared_attrs:
            raise AttributeError("{} has no attribute '{}'".format(self, item))

        return PathBuilder(model=self).extend_path(item=item)


class Model(with_metaclass(_ModelMeta)):
    """
    Declarative style models!

    Example usage:

        class SomeSalesforceObject(Model):
            # Optionally specify the name of the object as known by Salesforce.
            # This defaults to the class name.
            __salesforce_object_name__ = 'UglyAssName__c'

            id = attributes.String('Id')
            name = attributes.String('Name')

        instance = SomeSalesforceObject(id='123', name='BLAM!')
        assert instance.id == '123'
    """
    def __init__(self, **kwargs):
        """
        Constructs an instance of the Model.

        :param kwargs: attributes to set on the instance. These should map to
            the model attribute names, NOT the names as known by Salesforce.
        """
        self._instance_attrs = {}
        self._changed_attrs = set()

        # Crawl through the attributes declared in the class, pulling them
        # out of the provided kwargs and setting them as instance attributes.
        for key, attr in self.declared_attrs.items():
            if key not in kwargs:
                self._instance_attrs[key] = UNSET
            elif kwargs[key] is UNLOADED:
                self._instance_attrs[key] = UNLOADED
            else:
                self._instance_attrs[key] = attr.coerce(kwargs[key])

    def __getattr__(self, item):
        """
        Any attempts to retrieve to an attribute that doesn't seem to exist
        is redirected to the the model's mapped attributes.
        """
        if item not in self._instance_attrs:
            raise AttributeError("{} has no attribute '{}'".format(self, item))

        value = self._instance_attrs[item]
        if value is UNLOADED:
            raise AttributeNotLoaded(item)
        if value is UNSET:
            raise AttributeNotSet(item)
        return value

    def __setattr__(self, key, value):
        if key == '_instance_attrs' or key not in self._instance_attrs:
            super(Model, self).__setattr__(key, value)
        elif isinstance(self.__class__.declared_attrs[key], Relationship):
            super(Model, self).__setattr__(key, value)
        else:
            coerced = self.__class__.declared_attrs[key].coerce(value)
            if coerced != getattr(self, key):
                self._changed_attrs.add(key)

            self._instance_attrs[key] = coerced

    def __eq__(self, other):
        """
        Compares a model instance to another model instance.

        :param other: the other model instance being compared
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        return self._instance_attrs == other._instance_attrs

    @classmethod
    def __soql_name__(cls):
        """
        Returns the name of the model, as known by Salesforce.

        :rtype: str
        """
        return getattr(cls, '__salesforce_object_name__', cls.__name__)

    @classmethod
    def iter_columns(cls):
        """
        Iterates over the column attributes of a model (ignoring nasty things
        like relationships. Yuck.).
        """
        for key, attr in cls.declared_attrs.items():
            if isinstance(attr, Column):
                yield key

    @classmethod
    def load(cls, data):
        """
        Constructs an instance of the Model from Salesforce API data.

        :param dict data: dictionary of data coming straight from the
            Salesforce API.
        :rtype: admin_api_2.salesforce.orm.model.Model
        """
        attrs = {}
        for key, attr in cls.declared_attrs.items():
            if isinstance(attr, Column) and attr.salesforce_name not in data:
                # Hard fail if we were expecting a column from the API and
                # it didn't show up
                msg = "Expected {} for model {}. Didn't find that in: {}" \
                      "".format(attr.salesforce_name, cls, data)
                raise ExpectedColumnMissing(msg)
            elif isinstance(attr, Relationship) and attr.salesforce_name not in data:
                # Relationships, on the other hand, have to be explicitly
                # joined and its ok for them to be missing.
                attrs[key] = UNLOADED
            else:
                attrs[key] = attr.load(data[attr.salesforce_name])

        return cls(**attrs)

    def changes(self):
        model = self.__class__

        changes = {}
        for key in self._changed_attrs:
            attr = model.declared_attrs[key]
            value = attr.serialize(getattr(self, key))
            changes[attr.salesforce_name] = value

        return changes

    def reset_instance_state(self):
        self._changed_attrs = set()



from __future__ import absolute_import

from soql.attributes import (
    Boolean,
    Date,
    DateTime,
    Float,
    Integer,
    NullSalesforceColumnError,
    Relationship,
    String,
)

from soql.loaders import (
    load_model_from_salesforce_data,
    load_models_from_salesforce_data,
    get_total_count,
)

from soql.model import (
    AttributeNotLoaded,
    AttributeNotSet,
    ExpectedColumnMissing,
    Model,
)

from soql.model_registry import (
    ModelNotRegistered,
    ModelNotRegistered,
    model_registry
)

from soql.nodes import (
    asc,
    desc,
    nulls_first,
    nulls_last,
)

from soql.select import (
    select,
    SelectClauseIsntValidSubquery,
)



"""
    Utilities
    ~~~~~~~~~

    Random helpers.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
import sys


class AttrDict(dict):
    """This is just a little convenience class to make dealing with
    static sets prettier."""
    def __getattr__(self, attr):
        return self[attr]


if sys.version_info[0] < 3:
    to_unicode = unicode
else:
    to_unicode = str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest
from datetime import datetime, date

from dateutil.tz import tzutc

from soql import attributes
from soql import NullSalesforceColumnError
from soql import Model


class RelatedModel(Model):
    id = attributes.Integer('Id')


class RelationshipTest(unittest.TestCase):
    def test_load(self):
        attr = attributes.Relationship('Attr', related_model=RelatedModel)
        data = {
            'attributes': {'type': 'RelatedModel'},
            'Id': 1
        }
        self.assertEqual(attr.load(data), RelatedModel.load({'Id': 1}))

    def test_load_many(self):
        attr = attributes.Relationship('Attr', related_model=RelatedModel, many=True)
        data = {
            'records': [
                {'attributes': {'type': 'RelatedModel'}, 'Id': 1}
            ]
        }
        self.assertEqual(attr.load(data), [RelatedModel.load({'Id': 1})])

    def test_load_when_relationship_is_none(self):
        attr = attributes.Relationship('Attr', related_model=RelatedModel)
        self.assertEqual(attr.load(None), None)

    def test_load_many_when_relationship_is_none(self):
        attr = attributes.Relationship('Attr', related_model=RelatedModel, many=True)
        self.assertEqual(attr.load(None), [])


class StringTest(unittest.TestCase):
    def test_coerce(self):
        attr = attributes.String('Attr')
        self.assertEqual(attr.coerce(1), '1')
        self.assertEqual(attr.coerce('1'), '1')
        self.assertEqual(attr.coerce('È'), 'È')
        self.assertRaises(NullSalesforceColumnError, attr.coerce, None)

    def test_nullable(self):
        attr = attributes.String('Attr', nullable=True)
        self.assertIsNone(attr.coerce(None))

    def test_serialize(self):
        attr = attributes.String('Attr')
        self.assertEqual(attr.serialize('1'), '1')


class IntegerTest(unittest.TestCase):
    def test_coerce(self):
        attr = attributes.Integer('Attr')
        self.assertEqual(attr.coerce('1'), 1)
        self.assertEqual(attr.coerce(1), 1)
        self.assertRaises(NullSalesforceColumnError, attr.coerce, None)

    def test_nullable(self):
        attr = attributes.Integer('Attr', nullable=True)
        self.assertIsNone(attr.coerce(None))

    def test_serialize(self):
        attr = attributes.Integer('Attr')
        self.assertEqual(attr.serialize(1), 1)


class FloatTest(unittest.TestCase):
    def test_coerce(self):
        attr = attributes.Float('Attr')
        self.assertEqual(attr.coerce('1.0'), 1.0)
        self.assertEqual(attr.coerce(1.0), 1.0)

    def test_nullable(self):
        attr = attributes.Float('Attr', nullable=True)
        self.assertIsNone(attr.coerce(None))

    def test_serialize(self):
        attr = attributes.Float('Attr')
        self.assertEqual(attr.serialize(1.1), 1.1)


class BooleanTest(unittest.TestCase):
    def test_coerce(self):
        attr = attributes.Boolean('Attr')
        self.assertEqual(attr.coerce(True), True)
        self.assertEqual(attr.coerce(False), False)
        self.assertRaises(NullSalesforceColumnError, attr.coerce, None)

    def test_nullable(self):
        attr = attributes.Boolean('Attr', nullable=True)
        self.assertIsNone(attr.coerce(None))

    def test_serialize(self):
        attr = attributes.Boolean('Attr')
        self.assertEqual(attr.serialize(True), True)


class DateTimeTest(unittest.TestCase):
    def test_coerce(self):
        attr = attributes.DateTime('Attr')

        datetime_ = datetime(year=1990, month=10, day=19, tzinfo=tzutc())
        date_ = date(year=1990, month=10, day=19)

        self.assertEqual(attr.coerce(datetime_), datetime_)
        self.assertEqual(attr.coerce(datetime_.isoformat()), datetime_)
        self.assertEqual(attr.coerce(datetime_.replace(tzinfo=None)), datetime_)
        self.assertEqual(attr.coerce(date_), datetime_)

        tz_naive_string = datetime_.replace(tzinfo=None).isoformat()
        self.assertEqual(attr.coerce(tz_naive_string), datetime_)
        self.assertRaises(NullSalesforceColumnError, attr.coerce, None)

    def test_nullable(self):
        attr = attributes.DateTime('Attr', nullable=True)
        self.assertIsNone(attr.coerce(None))

    def test_serialize(self):
        datetime_ = datetime(year=1990, month=10, day=19, tzinfo=tzutc())

        attr = attributes.DateTime('Attr')
        self.assertEqual(attr.serialize(datetime_), datetime_.isoformat())


class DateTest(unittest.TestCase):
    def test_coerce(self):
        attr = attributes.Date('Attr')

        datetime_ = datetime(year=1990, month=10, day=19, hour=5, tzinfo=tzutc())
        date_ = date(year=1990, month=10, day=19)

        self.assertEqual(attr.coerce(date_), date_)
        self.assertEqual(attr.coerce(date_.isoformat()), date_)
        self.assertEqual(attr.coerce(datetime_), date_)
        self.assertRaises(NullSalesforceColumnError, attr.coerce, None)

    def test_nullable(self):
        attr = attributes.Date('Attr', nullable=True)
        self.assertIsNone(attr.coerce(None))

    def test_serialize(self):
        date_ = date(year=1990, month=10, day=19)

        attr = attributes.Date('Attr')
        self.assertEqual(attr.serialize(date_), date_.isoformat())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest

from soql.attributes import Integer, Relationship, String
from soql import Model
from soql import select
from soql import SelectClauseIsntValidSubquery
from soql import asc, desc, nulls_first, nulls_last
from tests.helpers import SoqlAssertions


class Grandparent(Model):
    id = Integer('Id')


class Parent(Model):
    id = Integer('Id')
    name = String('Name')
    age = Integer('Age')
    mom = Relationship('Mom', related_model=Grandparent)


class Child(Model):
    id = Integer('Id')
    name = String('Name')
    mom = Relationship('Mom', related_model=Parent)
    dad = Relationship('Dad', related_model=Parent)
    teacher = Relationship('Teacher', related_model='Teacher')


class Teacher(Model):
    id = Integer('Id')
    students = Relationship('Students', related_model=Child, many=True)


class SelectTest(unittest.TestCase, SoqlAssertions):
    def test_select(self):
        self.assertSoqlEqual(
            select(Child),
            "SELECT Child.Id, Child.Name "
            "FROM Child"
        )

    def test_joins(self):
        self.assertSoqlEqual(
            select(Child).join(Child.mom),
            "SELECT Child.Id, Child.Name, Child.Mom.Age, Child.Mom.Id, Child.Mom.Name "
            "FROM Child"
        )

        self.assertSoqlEqual(
            select(Teacher).join(Teacher.students),
            "SELECT Teacher.Id, (SELECT Child.Id, Child.Name FROM Teacher.Students) "
            "FROM Teacher"
        )

        self.assertSoqlEqual(
            select(Teacher).join(Teacher.students).join(Teacher.students.mom),
            "SELECT Teacher.Id, "
            "(SELECT Child.Id, Child.Name, Child.Mom.Age, Child.Mom.Id, Child.Mom.Name FROM Teacher.Students) "
            "FROM Teacher"
        )

        self.assertSoqlEqual(
            select(Teacher).join(Teacher.students.mom),
            "SELECT Teacher.Id, "
            "(SELECT Child.Id, Child.Name, Child.Mom.Age, Child.Mom.Id, Child.Mom.Name FROM Teacher.Students) "
            "FROM Teacher"
        )

        self.assertSoqlEqual(
            select(Child).join(Child.mom.mom),
            "SELECT Child.Id, Child.Name, Child.Mom.Age, "
            "Child.Mom.Id, Child.Mom.Name, Child.Mom.Mom.Id "
            "FROM Child"
        )

        self.assertSoqlEqual(
            select(Teacher).join(Teacher.students.mom).join(
                Teacher.students.dad),
            "SELECT Teacher.Id, "
            "(SELECT Child.Id, Child.Name, Child.Dad.Age, Child.Dad.Id, Child.Dad.Name, "
            "Child.Mom.Age, Child.Mom.Id, Child.Mom.Name FROM Teacher.Students) "
            "FROM Teacher"
        )

        self.assertSoqlEqual(
            select(Child).join(Child.teacher.students.mom),
            "SELECT Child.Id, Child.Name, Child.Teacher.Id, "
            "(SELECT Child.Id, Child.Name, Child.Mom.Age, Child.Mom.Id, "
            "Child.Mom.Name FROM Child.Teacher.Students) "
            "FROM Child"
        )

    def test_filters(self):
        self.assertSoqlEqual(
            select(Child).where(Child.id == '123'),
            "SELECT Child.Id, Child.Name "
            "FROM Child "
            "WHERE Child.Id = '123'"
        )

        self.assertSoqlEqual(
            select(Child).where(Child.id == '123').where(Child.name == 'Jill'),
            "SELECT Child.Id, Child.Name "
            "FROM Child "
            "WHERE Child.Id = '123' AND Child.Name = 'Jill'"
        )

        self.assertSoqlEqual(
            select(Child).where(Child.name == u'CATMONKÈ-123490'),
            u"SELECT Child.Id, Child.Name "
            u"FROM Child "
            u"WHERE Child.Name = 'CATMONKÈ-123490'"
        )

    def test_order_by(self):
        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age"
        )

        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age).order_by(Parent.id),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age, Parent.Id"
        )

        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age, direction=desc),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age DESC"
        )

        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age, direction=desc).order_by(Parent.id, direction=asc),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age DESC, Parent.Id ASC"
        )

        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age, direction=asc, nulls_position=nulls_first),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age ASC NULLS FIRST"
        )

        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age, direction=desc, nulls_position=nulls_last),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age DESC NULLS LAST"
        )

    def test_count(self):
        self.assertSoqlEqual(
            select(Child).count(),
            "SELECT COUNT() "
            "FROM Child"
        )

    def test_offset_and_limit(self):
        self.assertSoqlEqual(
            select(Child).limit(100),
            "SELECT Child.Id, Child.Name "
            "FROM Child "
            "LIMIT 100"
        )

        self.assertSoqlEqual(
            select(Child).offset(100),
            "SELECT Child.Id, Child.Name "
            "FROM Child "
            "OFFSET 100"
        )

        self.assertSoqlEqual(
            select(Parent).order_by(Parent.age).offset(100).limit(100),
            "SELECT Parent.Age, Parent.Id, Parent.Name "
            "FROM Parent "
            "ORDER BY Parent.Age "
            "LIMIT 100 "
            "OFFSET 100"
        )

    def test_override_columns(self):
        self.assertSoqlEqual(
            select(Parent).columns(Parent.id),
            "SELECT Parent.Id "
            "FROM Parent"
        )

        self.assertSoqlEqual(
            select(Parent).columns(Parent.id, Parent.name),
            "SELECT Parent.Id, Parent.Name "
            "FROM Parent"
        )

    def test_subquery(self):
        self.assertSoqlEqual(
            select(Parent).columns(Parent.id).subquery(),
            "(SELECT Parent.Id FROM Parent)"
        )

        subquery = select(Parent).columns(Parent.name).subquery()
        self.assertSoqlEqual(
            select(Child).where(Child.name.in_(subquery)),
            "SELECT Child.Id, Child.Name "
            "FROM Child "
            "WHERE Child.Name IN (SELECT Parent.Name FROM Parent)"
        )

        with self.assertRaises(SelectClauseIsntValidSubquery):
            select(Parent).offset(100).subquery()



"""
    Path Builder
    ~~~~~~~~~~~~

    Helper for building relationship queries.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from __future__ import absolute_import

from soql.attributes import Relationship
from soql.nodes import ColumnPath
from soql.utils import to_unicode


class PathBuilder(object):
    """
    Tracks chained model-relationship references and exposes helper methods
    for interpreting the chain.
    """
    def __init__(self, model, path=None):
        """
        El constructor.

        :param admin_api_2.salesforce.orm.model.Model model:
        :param list path: Optionally instantiate the instance with an existing
            path. This is useful when chaining relationships.
        """
        self.model = model
        self.path = path or []

    def _iterate_path(self, path_index):
        """
        Iterates through the path, starting from the base model to the last
        relationship reference.

        This yields tuples of the relationship's model and the attribute
        for the relationship.

        For example, if we had.....


            relationship = attributes.Relationship('Parent')

            class SomeModel(Model):
                id = attributes.String('Id')
                parent = relationship

            path_builder = PathBuilder(SomeModel)\
                .extend_path(parent)\
                .extend_path(parent)

        This would yield.....

            (SomeModel, relationship)
            (SomeModel, relationship)

        :param integer path_index: Only iterate up to this point in the path.
            Use -1 to iterate over the entire path.
        """
        if path_index == -1:
            # -1 + 1 = 0... but -1 means we want to iterate through the entire
            # path, so we'll check for this special case.
            inclusive_index = None
        else:
            inclusive_index = path_index + 1

        last_model = self.model

        for relationship in self.path[0:inclusive_index]:
            model_relationship = last_model.declared_attrs[relationship]
            last_model = model_relationship.related_model
            yield last_model, model_relationship

    def _build_soql_column_path(self):
        """This builds the SOQL representation of this path."""
        # The column path starts with the name of the original object
        path = [self.model.__soql_name__()]

        for _, attr in self._iterate_path(path_index=-1):
            path.append(attr.salesforce_name)

        return path

    def get_model_in_path(self, path_index):
        """Get the model of a certain point in the relationship path."""
        last_model = self.model
        for model, _ in self._iterate_path(path_index=path_index):
            last_model = model
        return last_model

    def get_relationship_attr_in_path(self, path_index):
        """Get the relationship attribute of a certain point in the
        relationship path."""
        last_model_relationship_attr = None
        for _, attr in self._iterate_path(path_index=path_index):
            last_model_relationship_attr = attr
        return last_model_relationship_attr

    def get_column_node(self):
        """Builds a Node representing this path."""
        return ColumnPath(*self._build_soql_column_path())

    def get_last_model_column_nodes(self):
        """Returns SOQL representations of all the columns for the final
        model in the relationship path."""
        last_model = self.get_model_in_path(path_index=-1)
        return [
            ColumnPath(self.extend_path(attr))
            for attr in last_model.iter_columns()
        ]

    def extend_path(self, item):
        """Extends the path with another reference.

        If that reference is another relationship, this returns a new instance.

        If that reference is column, we've reached the end of our path, and this
        returns a Colum node.
        """
        last_model = self.get_model_in_path(path_index=-1)
        attr = last_model.declared_attrs[item]

        if isinstance(attr, Relationship):
            # Looks like someone is requesting another relationship, so we'll
            # keep extending the chain!
            return PathBuilder(
                model=self.model,
                path=self.path + [item]
            )

        # Looks like someone is requesting a column, which is a leaf in a path,
        # so we'll return a column node.
        path = self._build_soql_column_path() + [attr.salesforce_name]
        return ColumnPath(*path)

    def __getattr__(self, item):
        """
        This little bit of magic let's us do things like...

            select(SomeModel).where(SomeModel.relationship.name == 'BLAM!')
        """
        return self.extend_path(item=item)

    def __unicode__(self):
        return to_unicode(self.get_column_node())

    def __str__(self):
        return str(self.get_column_node())



"""
    Loaders
    ~~~~~~~

    Utilities for loading models from Salesforce API data.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from __future__ import absolute_import

from soql.model_registry import model_registry


def load_model_from_salesforce_data(data):
    """
    Loads a dictionary representing a single Salesforce object into a model.

    :param dict data: data from the Salesforce API
    :rtype: admin_api_2.salesforce.orm.model.Model
    """
    object_type = data['attributes']['type']
    model = model_registry.get_by_salesforce_object_name(name=object_type)
    return model.load(data)


def load_models_from_salesforce_data(data):
    """
    Loads a dictionary representing multiple Salesforce objects into a models.

    :param dict data: data from the Salesforce API
    :rtype: list
    """
    return list(map(load_model_from_salesforce_data, data['records']))


def get_total_count(data):
    """
    Retrieves the total count from a Salesforce SOQL query.

    :param dict data: data from the Salesforce API
    :rtype: int
    """
    return data['totalSize']



"""
    Nodes
    ~~~~~

    Nodes for using composite pattern to implement a query builder.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""

from __future__ import absolute_import
from __future__ import unicode_literals
from datetime import datetime

import six
from dateutil.tz import tzutc

from soql.utils import AttrDict
from soql.utils import to_unicode


# The various operations you can perform in SOQL
OPS = AttrDict(
    AND='AND',
    OR='OR',
    NOT='NOT',
    EQ='=',
    LT='<',
    LTE='<=',
    GT='>',
    GTE='>=',
    NE='!=',
    IN='IN',
    NOT_IN='NOT IN',
    IS='IS',
    IS_NOT='IS NOT',
    LIKE='LIKE',
    INCLUDES='INCLUDES',
    EXCLUDES='EXCLUDES'
)

# Reserved SOQL words
WORDS = AttrDict(
    SELECT='SELECT',
    FROM='FROM',
    WHERE='WHERE',
    GROUP_BY='GROUP BY',
    ORDER_BY='ORDER BY',
    LIMIT='LIMIT',
    OFFSET='OFFSET',
    ASC='ASC',
    DESC='DESC',
    NULL='NULL',
    TRUE='TRUE',
    FALSE='FALSE',
    NULLS_FIRST='NULLS FIRST',
    NULLS_LAST='NULLS LAST'
)

# SOQL functions
FUNCTIONS = AttrDict(
    AVG='AVG',
    COUNT='COUNT',
    COUNT_DISTINCT='COUNT_DISTINCT',
    MIN='MIN',
    MAX='MAX',
    SUM='SUM'
)


def _stringify(value):
    """Converts a native Python object to its SOQL representation."""
    if isinstance(value, six.string_types):
        return "'{}'".format(value)
    if value is None:
        return WORDS.NULL
    if value is True:
        return WORDS.TRUE
    if value is False:
        return WORDS.FALSE
    if isinstance(value, datetime):
        if not value.tzinfo:
            return value.replace(tzinfo=tzutc()).isoformat()
        else:
            return value.isoformat()
    return str(value)


class Node(object):
    """Base node for building a graph representation of a SOQL query."""
    def __init__(self, child_nodes, sep=''):
        self.child_nodes = child_nodes
        self.sep = sep

    def _as_string(self, method):
        return self.sep.join([method(node) for node in self.child_nodes])

    def __unicode__(self):
        return to_unicode(self._as_string(to_unicode))

    def __str__(self):
        return str(self._as_string(str))


class Grouped(Node):
    def __init__(self, node, braces='()'):
        super(Grouped, self).__init__(child_nodes=[braces[0], node, braces[1]])


class Expression(Node):
    def __init__(self, lhs, op, rhs):
        child_nodes = [_stringify(lhs), op, _stringify(rhs)]
        super(Expression, self).__init__(child_nodes=child_nodes, sep=' ')


class CommaSeparated(Node):
    def __init__(self, items):
        super(CommaSeparated, self).__init__(child_nodes=items, sep=', ')


class AndSeparated(Node):
    def __init__(self, items):
        sep = ' {} '.format(OPS.AND)
        super(AndSeparated, self).__init__(child_nodes=items, sep=sep)


class Array(Node):
    def __init__(self, items):
        array = Grouped(node=CommaSeparated(items=items))
        super(Array, self).__init__(child_nodes=[array])


class Function(Node):
    def __init__(self, function_name, args=None):
        child_nodes = [function_name, Array(args or [])]
        super(Function, self).__init__(child_nodes=child_nodes, sep='')


class Count(Function):
    def __init__(self, field_name=None):
        args = [field_name] if field_name is not None else []
        super(Count, self).__init__(function_name=FUNCTIONS.COUNT, args=args)


class SelectClause(Node):
    def __init__(
            self,
            select_columns,
            from_table,
            where_conditions=None,
            order_bys=None,
            limit=None,
            offset=None
    ):
        nodes = [
            WORDS.SELECT,
            CommaSeparated(items=select_columns),
            WORDS.FROM,
            from_table
        ]

        if where_conditions:
            nodes.append(WORDS.WHERE)
            nodes.append(AndSeparated(items=where_conditions))

        if order_bys:
            nodes.append(WORDS.ORDER_BY)
            nodes.append(CommaSeparated(items=order_bys))

        if limit is not None:
            nodes.append(WORDS.LIMIT)
            nodes.append(Node(child_nodes=[limit]))

        if offset is not None:
            nodes.append(WORDS.OFFSET)
            nodes.append(Node(child_nodes=[offset]))

        super(SelectClause, self).__init__(child_nodes=nodes, sep=' ')


class OrderByClause(Node):
    def __init__(self, column, direction=None, nulls_position=None):
        child_nodes = [column]
        if direction:
            child_nodes.append(direction)
        if nulls_position:
            child_nodes.append(nulls_position)
        super(OrderByClause, self).__init__(child_nodes=child_nodes, sep=' ')


class SubqueryClause(Node):
    def __init__(
            self,
            select_columns,
            from_table,
            where_conditions=None
    ):
        query = SelectClause(
            select_columns=select_columns,
            from_table=from_table,
            where_conditions=where_conditions
        )
        grouped = Grouped(node=query)
        super(SubqueryClause, self).__init__(child_nodes=[grouped])


def _op_factory(op):
    """Quick little factory for a little bit of DRY when defining magic methods
    on Operatable."""
    def inner(self, rhs):
        return Expression(lhs=self, op=op, rhs=rhs)
    return inner


def _get_container_node(value):
    """Used for building container expressions - the input could either be
    a Python iterable, or another node."""
    if isinstance(value, Node):
        # This looks like a subquery!
        return value
    else:
        return Array(items=[_stringify(i) for i in value])


class Operatable(object):
    """A mixin to give a node operation functionality. I.e., this is where
    some of the magic happens, and allows for the nice filter syntax:

        SomeModel.name == 'John'
        SomeModel.id.in_([1, 2, 3])
        etc.
    """
    __eq__ = _op_factory(OPS.EQ)
    __ne__ = _op_factory(OPS.NE)
    __gt__ = _op_factory(OPS.GT)
    __ge__ = _op_factory(OPS.GTE)
    __lt__ = _op_factory(OPS.LT)
    __le__ = _op_factory(OPS.LTE)

    def like_(self, value):
        return Expression(lhs=self, op=OPS.LIKE, rhs=value)

    def in_(self, iterable):
        container = _get_container_node(value=iterable)
        return Expression(lhs=self, op=OPS.IN, rhs=container)

    def not_in_(self, iterable):
        container = _get_container_node(value=iterable)
        return Expression(lhs=self, op=OPS.NOT_IN, rhs=container)

    def includes_(self, iterable):
        container = _get_container_node(value=iterable)
        return Expression(lhs=self, op=OPS.INCLUDES, rhs=container)

    def excludes_(self, iterable):
        container = _get_container_node(value=iterable)
        return Expression(lhs=self, op=OPS.EXCLUDES, rhs=container)


class ColumnPath(Node, Operatable):
    def __init__(self, *elements):
        super(ColumnPath, self).__init__(child_nodes=elements, sep='.')


def _logical_expression(nodes, op):
    expression = Expression(lhs=nodes[0], op=op, rhs=nodes[1])
    for node in nodes[2:]:
        expression = Expression(lhs=expression, op=op, rhs=node)
    return Grouped(node=expression)


def and_(*nodes):
    return _logical_expression(nodes=nodes, op=OPS.AND)


def or_(*nodes):
    return _logical_expression(nodes=nodes, op=OPS.OR)


def not_(node):
    return Node(child_nodes=[OPS.NOT, node], sep=' ')


# These values are needed publicly, so we'll create some nice aliases.
asc = WORDS.ASC
desc = WORDS.DESC
nulls_first = WORDS.NULLS_FIRST
nulls_last = WORDS.NULLS_LAST





import unittest
from datetime import date, datetime

from dateutil.tz import tzutc

from soql import attributes
from soql import Model
from soql import AttributeNotSet
from soql import AttributeNotLoaded
from soql import ExpectedColumnMissing
from soql.model import UNLOADED
from soql.model import UNSET


class AParent(Model):
    id = attributes.Integer('Id')
    name = attributes.String('Name')
    age = attributes.Integer('Age')


class AChild(Model):
    id = attributes.Integer('Id')
    name = attributes.String('Name')
    mom = attributes.Relationship('Mom', related_model=AParent)


class AllZeTypes(Model):
    id = attributes.Integer('Id')
    name = attributes.String('Name')
    price = attributes.Float('Price')
    start_date = attributes.Date('StartDate')
    created_time = attributes.DateTime('CreatedTime')


class AllZeRelationships(Model):
    parent = attributes.Relationship('Parent', related_model=AParent)
    children = attributes.Relationship('Children', related_model=AChild, many=True)


class ModelTest(unittest.TestCase):
    def test_loading_from_salesforce_api_data(self):
        start_date = date(year=1990, month=10, day=19)
        created_time = datetime(year=1990, month=10, day=19, tzinfo=tzutc())

        instance = AllZeTypes.load({
            'attributes': {'type': 'AllZeTypes'},
            'Id': '123',
            'Name': 11,
            'Price': '3.50',
            'StartDate': str(start_date),
            'CreatedTime': created_time.isoformat()
        })
        self.assertEqual(instance.id, 123)
        self.assertEqual(instance.name, '11')
        self.assertEqual(instance.price, 3.50)
        self.assertEqual(instance.start_date, start_date)
        self.assertEqual(instance.created_time, created_time)

        instance = AllZeRelationships.load({
            'Children': {
                'records': [
                    {'attributes': {'type': 'AChild'}, 'Id': 123, 'Name': 'Jack'},
                    {'attributes': {'type': 'AChild'}, 'Id': 456, 'Name': 'Jill'}
                ]
            }
        })
        self.assertEqual(instance.children, [
            AChild(id=123, name='Jack', mom=UNLOADED),
            AChild(id=456, name='Jill', mom=UNLOADED)
        ])
        self.assertRaises(AttributeNotLoaded, lambda: instance.parent)

        instance = AllZeRelationships.load({
            'Parent': {
                'attributes': {'type': 'AParent'},
                'Id': 123,
                'Name': 'Jack',
                'Age': 45}
        })
        self.assertEqual(instance.parent, AParent(id=123, name='Jack', age=45))
        self.assertRaises(AttributeNotLoaded, lambda: instance.children)

        # Assert loading hard fails if a column is missing
        with self.assertRaises(ExpectedColumnMissing):
            instance = AllZeTypes.load({
                'attributes': {'type': 'AllZeTypes'},
                'Id': '123',
                'Name': 11,
            })

    def test_instantiation(self):
        start_date = date(year=1990, month=10, day=19)
        created_time = datetime(year=1990, month=10, day=19, tzinfo=tzutc())

        instance = AllZeTypes(
            id=123,
            name='11',
            price=3.50,
            start_date=start_date,
            created_time=created_time
        )
        self.assertEqual(instance.id, 123)
        self.assertEqual(instance.name, '11')
        self.assertEqual(instance.price, 3.50)
        self.assertEqual(instance.start_date, start_date)
        self.assertEqual(instance.created_time, created_time)

        instance = AllZeRelationships(
            children=[AChild(id=123, name='Jack')]
        )
        self.assertEqual(instance.children, [
            AChild(id=123, name='Jack', mom=UNSET)
        ])
        self.assertRaises(AttributeNotSet, lambda: instance.parent)

        instance = AllZeRelationships(
            parent=AParent(id=123, name='Jack', age=45)
        )
        self.assertEqual(instance.parent, AParent(id=123, name='Jack', age=45))
        self.assertRaises(AttributeNotSet, lambda: instance.children)

    def test_multiple_inheritance(self):
        class Preferences(Model):
            candy = attributes.String('Candy')

        class FavoriteChild(AChild, Preferences):
            pass

        favorite_child = FavoriteChild(
            id=123,
            name='Jennifer',
            candy='twizzlers'
        )

        self.assertEqual(favorite_child.id, 123)
        self.assertEqual(favorite_child.name, 'Jennifer')
        self.assertEqual(favorite_child.candy, 'twizzlers')

    def test_back_references(self):
        class Foo(Model):
            bar = attributes.Relationship('Bar', related_model='Bar')

        class Bar(Model):
            foos = attributes.Relationship('Foos', related_model=Foo, many=True)

        bar = Bar()
        foo = Foo(bar=bar)
        self.assertEqual(foo.bar, bar)

        foos = [Foo() for _ in range(3)]
        bar = Bar(foos=foos)
        self.assertEqual(bar.foos, foos)

    def test_set_attributes(self):
        child = AChild(id=456, name='Jill')
        child.name = 'MOTHA FUCKIN JILL'
        self.assertEqual(child.name, 'MOTHA FUCKIN JILL')

    def test_changes(self):
        start_date = date(year=1990, month=10, day=19)
        new_date = date(year=1990, month=10, day=20)

        instance = AllZeTypes(id=456, name='Jill', start_date=start_date)

        self.assertEqual(instance.changes(), {})

        instance.name = 'Bill'

        self.assertEqual(instance.changes(), {
            'Name': 'Bill'
        })

        # Test that the value is coerced
        instance.start_date = new_date

        self.assertEqual(instance.changes(), {
            'StartDate': new_date.isoformat(),
            'Name': 'Bill'
        })

        instance.reset_instance_state()

        self.assertEqual(instance.changes(), {})

        # No changes!
        instance.name = 'Bill'

        self.assertEqual(instance.changes(), {})

        instance.name = 'Jill'

        self.assertEqual(instance.changes(), {
            'Name': 'Jill'
        })



import unittest

from soql import attributes
from soql import Model
from soql import ModelNotRegistered
from soql import load_model_from_salesforce_data
from soql import load_models_from_salesforce_data
from soql import get_total_count


class ModelBase(Model):
    __abstract__ = True
    id = attributes.Integer('Id')


class LoadedToMany(ModelBase):
    __salesforce_object_name = 'LoadedToMany__c'


class LoadedToOne(ModelBase):
    __salesforce_object_name__ = 'LoadedToOne__c'


class Loaded(ModelBase):
    id = attributes.Integer('Id')
    one = attributes.Relationship('One', related_model=LoadedToOne)
    many = attributes.Relationship('Many', related_model=LoadedToMany, many=True)


class LoaderTest(unittest.TestCase):
    def format_dict_as_api_data(self, data, object_type):
        formatted = {
            'attributes': {'type': object_type},
        }
        formatted.update(data)
        return formatted

    def format_collection_as_api_data(self, collection, object_type):
        formatted = {
            'totalSize': len(collection),
            'records': [
                self.format_dict_as_api_data(data=o, object_type=object_type)
                for o in collection
            ]
        }
        return formatted

    def test_load_model_from_salesforce_data(self):
        one = self.format_dict_as_api_data(
            data={'Id': 1},
            object_type=LoadedToOne.__soql_name__()
        )
        many = self.format_collection_as_api_data(
            collection=[{'Id': 2}, {'Id': 3}],
            object_type=LoadedToMany.__soql_name__()
        )
        data = self.format_dict_as_api_data(
            data={'Id': 4, 'One': one, 'Many': many},
            object_type=Loaded.__soql_name__()
        )
        loaded = load_model_from_salesforce_data(data=data)

        self.assertEqual(
            loaded,
            Loaded(
                id=4,
                one=LoadedToOne(id=1),
                many=[LoadedToMany(id=2), LoadedToMany(id=3)]
            )
        )

    def test_load_models_from_salesforce_data(self):
        one = self.format_dict_as_api_data(
            data={'Id': 1},
            object_type=LoadedToOne.__soql_name__()
        )
        many = self.format_collection_as_api_data(
            collection=[{'Id': 2}, {'Id': 3}],
            object_type=LoadedToMany.__soql_name__()
        )
        data = self.format_collection_as_api_data(
            collection=[{'Id': 4, 'One': one, 'Many': many}],
            object_type=Loaded.__soql_name__()
        )
        loaded = load_models_from_salesforce_data(data=data)

        self.assertEqual(
            loaded,
            [
                Loaded(
                    id=4,
                    one=LoadedToOne(id=1),
                    many=[LoadedToMany(id=2), LoadedToMany(id=3)]
                )
            ]
        )

    def test_get_total_count(self):
        data = self.format_collection_as_api_data(
            collection=[{'Id': 1}, {'Id': 2}],
            object_type=Loaded.__soql_name__()
        )
        self.assertEqual(get_total_count(data=data), 2)

    def test_abstract_models_are_ignored(self):
        data = self.format_dict_as_api_data(
            data={'Id': 1},
            object_type=ModelBase.__soql_name__()
        )

        with self.assertRaises(ModelNotRegistered):
            load_model_from_salesforce_data(data=data)



"""
    Attributes
    ~~~~~~~~~~

    Attributes for declarative models.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from __future__ import absolute_import
from datetime import datetime, date

from dateutil.parser import parse
from dateutil.tz import tzutc

from soql.model_registry import model_registry
from soql.loaders import load_models_from_salesforce_data
from soql.loaders import load_model_from_salesforce_data


class NullSalesforceColumnError(Exception):
    def __init__(self, attribute):
        self.attribute = attribute
        msg = '{attribute} is unexpectedly null'.format(
            attribute=attribute.salesforce_name,
        )
        super(NullSalesforceColumnError, self).__init__(msg)


class AttributeBase(object):
    def __init__(self, salesforce_name, nullable=False):
        """
        Base Attribute class for declarative models.

        :param str salesforce_name: This should be the attribute name as it
            is in Salesforce. The ORM will use this when loading data from
            the API and for compiling SOQL statements.
        :param bool nullable: If True, this field can be None.
        """
        self.salesforce_name = salesforce_name
        self.nullable = nullable

    def load(self, value):
        """
        This is called when loading data from the Salesforce API, and can be
        used to pre process data before coercion.

        :param value: the attribute value coming from Salesforce (note - this
            value will have already been JSON loaded)
        :return: the processed value.
        """
        return value

    def coerce(self, value):
        """
        This is called on model instantiation and is used to coerce the value
        to the attribute's type.

        :param value: the pre-processed attribute value.
        :return: the coerced value.
        """
        if self.nullable and value is None:
            return value
        if not self.nullable and value is None:
            raise NullSalesforceColumnError(attribute=self)
        else:
            return self._coerce(value=value)

    def _coerce(self, value):
        return value


class Relationship(AttributeBase):
    def __init__(self, salesforce_name, related_model, many=False):
        """
        This special attribute represents a relationship to another model.

        :param salesforce_name:
        :param admin_api_2.salesforce.orm.model.Model|str related_model: The
            object type this is a relationship to. The Model itself or a string
            can be used - the latter of which allows for circular references.
        :param many: If False (default), this represents a many-to-one
            relationship. If True, this represents a one-to-many relationship.
        """
        super(Relationship, self).__init__(salesforce_name=salesforce_name)
        self._related_model = related_model
        self.many = many

    @property
    def related_model(self):
        if isinstance(self._related_model, str):
            # Looks like the related model is a string.... that's ok, as it
            # allows for back references. Fetch the model from the registry.
            return model_registry.get_by_model_name(self._related_model)
        return self._related_model

    def load(self, value):
        if self.many and value is None:
            return []
        if self.many:
            return load_models_from_salesforce_data(data=value)
        if value is None:
            return None
        return load_model_from_salesforce_data(data=value)


class Column(AttributeBase):
    def serialize(self, value):
        return value


class String(Column):
    def _coerce(self, value):
        try:
            return str(value)
        except UnicodeEncodeError:
            return value


class Integer(Column):
    def _coerce(self, value):
        return int(value)


class Float(Column):
    def _coerce(self, value):
        return float(value)


class Boolean(Column):
    def _coerce(self, value):
        return bool(value)


class Date(Column):
    def _coerce(self, value):
        if isinstance(value, datetime):
            return date(
                year=value.year,
                month=value.month,
                day=value.day
            )
        if isinstance(value, date):
            return value
        else:
            dt = parse(value)
            return date(
                year=dt.year,
                month=dt.month,
                day=dt.day
            )

    def serialize(self, value):
        return value.isoformat()


class DateTime(Column):
    def _coerce(self, value):
        if isinstance(value, datetime):
            if not value.tzinfo:
                # Assume timezone naive datetimes are UTC
                return value.replace(tzinfo=tzutc())
            return value
        if isinstance(value, date):
            return datetime(
                year=value.year,
                month=value.month,
                day=value.day,
                tzinfo=tzutc()
            )
        else:
            return self.coerce(parse(value))

    def serialize(self, value):
        return value.isoformat()



"""
    Select
    ~~~~~~

    Entrypoint for generating SOQL from models.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""

from __future__ import absolute_import
from collections import OrderedDict
from copy import copy

from soql.nodes import SubqueryClause
from soql.nodes import OrderByClause
from soql.nodes import Count
from soql.nodes import SelectClause
from soql.path_builder import PathBuilder
from soql.utils import to_unicode


class SelectClauseIsntValidSubquery(Exception):
    pass


class RelationshipNode(object):
    """
    Joins can be recursive. This little class let's us represent all the joined
    relationships of a select query as a graph.
    """
    def __init__(self, relationship=None):
        """
        Constructs a new node in a graph.

        :param str|NoneType relationship: The attribute name of the
            relationship. None represents the root node.
        """
        self.relationship = relationship

        # It helps testability tremendously if the joins come in a
        # reliable order... hence the OrderedDict
        self.child_relationships = OrderedDict()

    def __copy__(self):
        node = RelationshipNode(relationship=self.relationship)
        node.child_relationships = copy(self.child_relationships)
        return node

    def __iter__(self):
        for value in self.child_relationships.values():
            yield value

    def add_child(self, relationship):
        """Adds a child node."""
        self.child_relationships[relationship] = RelationshipNode(relationship)

    def add_path(self, path):
        """Adds the complete path to this node, creating any child nodes that
        don't exist yet."""
        node = self
        for relationship in path:
            if relationship not in node.child_relationships:
                node.add_child(relationship)
            node = node.child_relationships[relationship]


class SelectClauseBuilder(object):
    """
    Builds SOQL SELECT statements.
    """
    def __init__(self, model):
        """
        The model being selected from.

        :param admin_api_2.salesforce.orm.model.Model model:
        """
        self._model = model
        self._columns = PathBuilder(model=self._model).get_last_model_column_nodes()
        self._relationship_graph = RelationshipNode()
        self._order_bys = []
        self._filters = []
        self._limit = None
        self._offset = None
        self._count = False

    def __copy__(self):
        builder = SelectClauseBuilder(model=self._model)
        builder._columns = copy(self._columns)
        builder._filters = copy(self._filters)
        builder._order_bys = copy(self._order_bys)
        builder._relationship_graph = copy(self._relationship_graph)
        builder._limit = self._limit
        builder._offset = self._offset
        builder._count = self._count
        return builder

    def join(self, path_builder):
        """
        Add a join to the clause.

        :param path_builder:
        :rtype: SelectClauseBuilder
        """
        self._relationship_graph.add_path(path_builder.path)
        return copy(self)

    def where(self, expression):
        """
        Add a filter to the clause.

        :param expression:
        :rtype: SelectClauseBuilder
        """
        self._filters.append(expression)
        return copy(self)

    def order_by(self, column, direction=None, nulls_position=None):
        """
        Order the results of the clause.

        :param column:
        :param str direction: optionally declare the direction to order.
        :param str nulls_position: optionally declare whether the null values
            should be at the top or bottom of the results.
        :rtype: SelectClauseBuilder
        """
        clause = OrderByClause(
            column=column,
            direction=direction,
            nulls_position=nulls_position
        )
        self._order_bys.append(clause)

        return copy(self)

    def limit(self, limit):
        """
        Limit the amount of results.

        :param limit:
        :rtype: SelectClauseBuilder
        """
        self._limit = limit
        return copy(self)

    def offset(self, offset):
        """
        Skip the first amount of results.

        :param offset:
        :rtype: SelectClauseBuilder
        """
        self._offset = offset
        return copy(self)

    def count(self):
        """
        Instead of fetching the results, just count the total.

        :rtype: SelectClauseBuilder
        """
        self._count = True
        return copy(self)

    def columns(self, *columns):
        """
        Override the columns that are selected.

        :param columns:
        :rtype: SelectClauseBuilder
        """
        self._columns = list(columns)
        return copy(self)

    def subquery(self):
        """
        Convert the clause to a subquery. This is useful for things like...

            subquery = select(SomeModel.id).where(SomeModel.id == 123).subquery()

            select(SomeOtherModel).where(SomeOtherModel.fk.in_(subquery))

        :rtype: admin_api_2.salesforce.orm.soql.SubqueryClause
        """
        if self._order_bys or self._limit or self._offset:
            raise SelectClauseIsntValidSubquery()

        return SubqueryClause(
            select_columns=self._get_columns(),
            from_table=self._model.__soql_name__(),
            where_conditions=self._filters
        )

    def __unicode__(self):
        """Converts the instance to a SOQL string."""
        return to_unicode(self._get_select_clause())

    def __str__(self):
        return str(self._get_select_clause())

    def _get_columns(self):
        if self._count:
            return [Count()]

        joins = self._compile_joins()
        return self._columns + joins

    def _get_select_clause(self):
        return SelectClause(
            select_columns=self._get_columns(),
            from_table=self._model.__soql_name__(),
            where_conditions=self._filters,
            order_bys=self._order_bys,
            limit=self._limit,
            offset=self._offset
        )

    def _compile_joins(self):
        return self._recursively_compile_joins(
            path=PathBuilder(model=self._model),
            node=self._relationship_graph
        )

    def _recursively_compile_joins(self, path, node):
        """
        Depth-first-searches the relationship graph to build a SOQL
        representation of the joins.

        :param admin_api_2.salesforce.orm.path_builder.PathBuilder path:
        :param RelationshipNode node:
        :return: list of join clauses
        """
        joins = []

        for child_node in node:
            child_path = path.extend_path(child_node.relationship)
            attr = child_path.get_relationship_attr_in_path(path_index=-1)
            if attr.many:
                # Joining a one-to-many relationship needs to be done as a
                # subquery, where the subquery is scoped to the object being
                # joined.
                aggregate_join = self._recursively_compile_joins(
                    path=PathBuilder(model=attr.related_model),
                    node=child_node
                )
                from_table = child_path.get_column_node()
                subquery = SubqueryClause(
                    select_columns=aggregate_join,
                    from_table=from_table,
                )
                joins.append(subquery)

            else:
                column_joins = self._recursively_compile_joins(
                    path=child_path,
                    node=child_node
                )

                # The joined columns need to come before any subqueries
                joins = column_joins + joins

        if node.relationship is None:
            # Looks like we're back at the root node.
            return joins

        joins = path.get_last_model_column_nodes() + joins
        return joins


select = SelectClauseBuilder



"""
    Model Registry
    ~~~~~~~~~~~~~~

    Global registry for models.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""

class ModelAlreadyRegistered(Exception):
    pass


class ModelNotRegistered(Exception):
    pass


class ModelRegistry(object):
    """A simple store for all models that have been declared."""

    def __init__(self):
        # To avoid O(n) lookups, We'll maintain a mapping from the model's
        # class name to the model AND a mapping from the model's associated
        # Saleforce object name to the model.
        self._registry_by_sf_obj_name = {}
        self._registry_by_model_name = {}

    def add(self, model):
        """
        Adds a model to the store.

        :param admin_api_2.salesforce.orm.model.Model model:
        """
        salesforce_object_name = model.__soql_name__()
        model_name = model.__name__

        if salesforce_object_name in self._registry_by_sf_obj_name:
            raise ModelAlreadyRegistered(
                'Salesforce object name: {}'.format(salesforce_object_name)
            )
        if model_name in self._registry_by_model_name:
            raise ModelAlreadyRegistered(
                'Model name: {}'.format(model_name)
            )

        self._registry_by_sf_obj_name[salesforce_object_name] = model
        self._registry_by_model_name[model_name] = model

    def get_by_salesforce_object_name(self, name):
        """
        Retrieves a model from the registry with the object's name as known
        by Salesforce.

        :param str name:
        :rtype: admin_api_2.salesforce.orm.model.Model
        """
        if name not in self._registry_by_sf_obj_name:
            raise ModelNotRegistered('Salesforce object name: {}'.format(name))
        return self._registry_by_sf_obj_name[name]

    def get_by_model_name(self, name):
        """
        Retrieves a model from the registry with model's class name.

        :param str name:
        :rtype: admin_api_2.salesforce.orm.model.Model
        """
        if name not in self._registry_by_model_name:
            raise ModelNotRegistered('Model name: {}'.format(name))
        return self._registry_by_model_name[name]


model_registry = ModelRegistry()



#!/usr/bin/env python
from setuptools import setup


if __name__ == '__main__':
    setup(
        name='soql',
        version='1.1.0',
        author='Barak Alon',
        author_email='barak.s.alon@gmail.com',
        description='Models and query generator for Salesforce Object Query Language (SOQL)',
        long_description=open('README.rst').read(),
        keywords=['salesforce', 'soql', 'salesforce.com'],
        license='MIT',
        packages=['soql'],
        install_requires=[
            'python-dateutil',
            'six',
        ],
        zip_safe=True,
        url='https://github.com/plangrid/soql',
        classifiers=[
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ]
    )



import unittest

from soql import attributes
from soql import Model
from soql.path_builder import PathBuilder
from tests.helpers import SoqlAssertions


# Declare the relationships separately so we can assert some stuff in the tests
# more easily...
next_ = attributes.Relationship('Next', related_model='LinkedList')
list_ = attributes.Relationship('List', related_model='LinkedList')
children = attributes.Relationship('Children', related_model='GraphNode', many=True)


class LinkedList(Model):
    id = attributes.Integer('Id')
    next = next_


class GraphNode(Model):
    id = attributes.Integer('Id')
    value = attributes.Integer('Value')
    list = list_
    children = children


class PathBuilderTest(unittest.TestCase, SoqlAssertions):
    def assertColumnNodesEqual(self, node_list, expected):
        self.assertEqual(
            sorted([str(node) for node in node_list]),
            sorted(expected)
        )

    def test_getattr(self):
        self.assertSoqlEqual(LinkedList.id, 'LinkedList.Id')
        self.assertSoqlEqual(LinkedList.next.id, 'LinkedList.Next.Id')
        self.assertSoqlEqual(LinkedList.next.next.id, 'LinkedList.Next.Next.Id')
        self.assertSoqlEqual(GraphNode.children.list.next.id, 'GraphNode.Children.List.Next.Id')
        self.assertSoqlEqual(GraphNode.children.children, 'GraphNode.Children.Children')

    def test_get_relationship_attr_in_path(self):
        path = GraphNode.children.list.next
        self.assertEqual(path.get_relationship_attr_in_path(path_index=0), children)
        self.assertEqual(path.get_relationship_attr_in_path(path_index=1), list_)
        self.assertEqual(path.get_relationship_attr_in_path(path_index=2), next_)
        self.assertEqual(path.get_relationship_attr_in_path(path_index=-1), next_)

    def test_get_model_in_path(self):
        path = GraphNode.children.list.next
        self.assertEqual(path.get_model_in_path(path_index=0), GraphNode)
        self.assertEqual(path.get_model_in_path(path_index=1), LinkedList)
        self.assertEqual(path.get_model_in_path(path_index=2), LinkedList)
        self.assertEqual(path.get_model_in_path(path_index=-1), LinkedList)

    def test_get_last_model_column_nodes(self):
        path_builder = PathBuilder(model=GraphNode)
        self.assertColumnNodesEqual(
            path_builder.get_last_model_column_nodes(),
            ['GraphNode.Id', 'GraphNode.Value']
        )

        path_builder = path_builder.extend_path('children')
        self.assertColumnNodesEqual(
            path_builder.get_last_model_column_nodes(),
            ['GraphNode.Children.Id', 'GraphNode.Children.Value']
        )

        path_builder = path_builder.extend_path('list')
        self.assertColumnNodesEqual(
            path_builder.get_last_model_column_nodes(),
            ['GraphNode.Children.List.Id']
        )

